#ifndef persistence_h_aPeinoXeiFeethuz
#define persistence_h_aPeinoXeiFeethuz

#include <iostream>
#include <ostream>
#include <vector>

#include "simplicialcomplex.h"
#include "booleanmatrix.h"


namespace libstick {


template<class T>
std::ostream& operator<<(std::ostream &, const std::vector<T> &);


/** Persistence computers various persistent homology information on a
 * simplex_order of a simplicial_complex. */
template<int MAXDIM, class IT, class VT>
class persistence {

    public:
        typedef typename simplicial_complex<MAXDIM, IT, VT>::simplex_order simplex_order;
        typedef typename simplex_order::boundary_matrix boundary_matrix;
        typedef boolean_colmatrix<IT> transformation_matrix;

        /** A point in a diagram comprises birth- and death-index. These are the
         * indices of the simplices of 'order' that give birth and death to the
         * class. */
        struct diagram_point {
            IT birth;
            IT death;
        };


    public:
        /** Create a new peristence object on the given simplex_order */
        persistence(const simplex_order &order) :
            order(order),
            bm(0),
            rbm(0),
            tm(0)
        {
            reset();
        }

        /** Reset all results gained from 'order'. */
        void reset() {
            done_matrices = false;
        }

        /** Get boundary matrix 'bm' of 'order', compute reduces boundary
         * matrix 'rbm', and the transformation matrix 'tm' such that rbm = bm *
         * tm. */
        void compute_matrices() {
            if (done_matrices)
                return;
            done_matrices = true;

            bm = order.get_boundary_matrix();
            rbm = bm;
            tm = create_unit_matrix<transformation_matrix>(bm.size());
            reduce_boundary_matrix(rbm, tm);

            assert(rbm == bm * tm);
        }

        /** Get the boundary matrix of 'order'. */
        const boundary_matrix& get_boundary_matrix() {
            compute_matrices();
            return bm;
        }

        /** Get the reduced boundary matrix of 'order'. */
        const boundary_matrix& get_reduced_boundary_matrix() {
            compute_matrices();
            return rbm;
        }

        /** Get the transformation matrix of 'order', which transforms the
         * boundary matrix to the reduced boundary matrix, when multiplied from
         * the right. */
        const transformation_matrix& get_transformation_matrix() {
            compute_matrices();
            return tm;
        }


        /** Print the sequence of births and deaths of cycles (homology classes). */
        void interpret_reduction(std::ostream &os) {
            compute_matrices();
            for (unsigned c=0; c < bm.size(); ++c) {
                os << c << ". inserting ";

                switch (bm.get_column(c).size()) {
                    case 0:
                        os << "dummy vertex of dim -1";
                        break;
                    case 1:
                        os << "vertex";
                        break;
                    case 2:
                        os << "edge";
                        break;
                    case 3:
                        os << "triangle";
                        break;
                    case 4:
                        os << "tetrahedron";
                        break;
                    default:
                        os << "simplex";
                        break;
                }

                os << std::endl;
                os << "  boundary: " << bm.get_column(c) << std::endl;

                typename boolean_colrowmatrix<IT>::colbase::column_type col = rbm.get_column(c);
                if (col.size() == 0) {
                    os << "  \033[1;32mbirth\033[0;m of a cycle: " << tm.get_column(c) << std::endl;
                } else {
                    os << "  \033[1;31mdeath\033[0;m of a cycle: " << rbm.get_column(c) << std::endl;
                    os << "  boundary of: " << tm.get_column(c) << std::endl;
                }
                os << std::endl;
            }
        }

    private:

        /** The underlying simplex order of this diagram. */
        const simplex_order &order;

        bool done_matrices;

        boundary_matrix bm;
        boundary_matrix rbm;
        transformation_matrix tm;
};


/** Takes the boundary matrix b and transforms it inplace into a reduced matrix
 * b. The matrix v is required to be a unit matrix having the same dimensions
 * as b. It is maintained in order to leave the product b*inverse(v) as an
 * invariant.  Hence, the resulting b is equal to the product of the boundary
 * matrix times v. */
template<class IT, class D>
void reduce_boundary_matrix(boolean_colrowmatrix<IT> &b, boolean_colmatrix_base<IT, D> &v) {
    assert(b.size() == v.width());

    // Make every column reduced, i.e., it is a zero-column or contains only one 1.
    for (unsigned c=0; c < b.size(); ++c) {
        const typename boolean_colrowmatrix<IT>::colbase::column_type &col = b.get_column(c);
        if (col.size() == 0)
            continue;

        // The column is not a zero-column, take the lowest 1, and reduce the other columns at the row
        IT r = col.back();
        assert(b.get(r, c));

        // Get a copy of the r-th row
        typename boolean_colrowmatrix<IT>::rowbase::row_type row(b.get_row(r).size());
        copy(b.get_row(r).begin(), b.get_row(r).end(), row.begin());
        assert(row.size() >= 1);

        // Get rid of 1s at that row right of column c.
        typename boolean_colrowmatrix<IT>::row_type::const_iterator it = lower_bound(row.begin(), row.end(), c);
        for(; it != row.end(); ++it) {
            if (*it <= c)
                continue;

            assert(b.get(r, *it));
            b.add_column(*it, col);
            v.add_column(*it, v.get_column(c));
            assert(!b.get(r, *it));
        }
    }
}

template<class T>
std::ostream& operator<<(std::ostream& os, const std::vector<T> &vec) {
    os << "[";

    typename std::vector<T>::const_iterator it = vec.begin();
    while ( it != vec.end()) {
        os << *it;
        if (++it != vec.end())
            os << " ";
    }

    os << "]";
    return os;
}


} // namespace libstick

#endif
